/**
  *  \file ptrvector.h
  *  \brief Template class ptr_vector
  */
#ifndef PCC_V2_PTRVECTOR_H
#define PCC_V2_PTRVECTOR_H

#include <vector>
#include <utility>
#include <iterator>
#include "cpluslib/assert.h"

/** Vector of Pointers. Implements ownership semantics for a vector of
    pointers, and attempts to manage it in a type- and exception-safe way.

    The basic assumption is that ptr_vector<> owns objects passed to it.
    It will delete them when the slot holding the pointer goes away. You
    must not delete a pointer obtained from ptr_vector<>, or place it in
    another ptr_vector<>.

    Otherwise, the interface attempts to be somewhat close to
    std::vector<>, restricted to operations that are safe enough to
    maintain ownership semantics. */
template<typename T>
class ptr_vector {
    ptr_vector& operator=(const ptr_vector&);
    ptr_vector(const ptr_vector&);

    /** Content. Implemented as vector<void*>, so all instantiations share
        their delete, resize, etc. implementations. */
    std::vector<void*> data;
    void delete_all();

 public:
    class iterator;
    typedef iterator const_iterator;
    typedef typename std::vector<void*>::size_type size_type;
    typedef typename std::vector<void*>::difference_type difference_type;
    typedef T* value_type;

    ptr_vector();
    ~ptr_vector();

    T* push_back(T* t);
    void pop_back();
    void swap(ptr_vector& other);
    size_type size() const;
    bool empty() const;
    void clear();
    T* operator[](size_type index) const;
    T* front() const;
    T* back() const;

    const_iterator begin() const;
    const_iterator end() const;

    void reserve(size_type size);

    void erase(iterator i);
    void erase(iterator i, iterator j);

    /* ptr_vector-specific functions: */
    void swap_elems(size_type a, size_type b);
    void swap_elems(size_type a, ptr_vector& other, size_type b);
    void replace_elem(size_type a, T* value);
    T*   extract_elem(size_type a);
    T*   extract_last();
};

/** Random-access iterator. Supports
    - post/pre increment/decrement
    - addition and subtraction of difference_type
    - subtraction of another iterator
    - indexing
    - dereferencing
    - comparison */
template<typename T>
class ptr_vector<T>::iterator
    : public std::iterator<std::random_access_iterator_tag,
                           const T*,
                           std::vector<void*>::difference_type,
                           const T*,
                           const T*&>
{
    std::vector<void*>::const_iterator it;
 public:
    typedef std::vector<void*>::difference_type difference_type;

    // Constructor
    iterator(std::vector<void*>::const_iterator it)
        : it(it) { }

    // Increment and Decrement
    iterator& operator++()
        { ++it; return *this; }
    iterator operator++(int)
        { iterator res(it); ++it; return res; }
    iterator& operator--()
        { --it; return *this; }
    iterator operator--(int)
        { iterator res(it); --it; return res; }

    // Add and subtract
    iterator operator+(difference_type n) const
        { return iterator(it + n); }
    iterator operator-(difference_type n) const
        { return iterator(it - n); }

    // Advance
    iterator& operator+=(difference_type n)
        { it += n; return *this; }
    iterator& operator-=(difference_type n)
        { it -= n; return *this; }

    // Distance
    difference_type operator-(iterator rhs) const
        { return it - rhs.it; }

    // Comparison
    bool operator<(iterator rhs) const
        { return it < rhs.it; }
    bool operator>(iterator rhs) const
        { return it > rhs.it; }
    bool operator<=(iterator rhs) const
        { return it <= rhs.it; }
    bool operator>=(iterator rhs) const
        { return it >= rhs.it; }
    bool operator==(iterator rhs) const
        { return it == rhs.it; }
    bool operator!=(iterator rhs) const
        { return it != rhs.it; }

    // Dereferencing and indexing
    T* operator*() const
        { return static_cast<T*>(*it); }
    T* operator[](int n) const
        { return static_cast<T*>(it[n]); }
};

/** Constructor. Makes a blank vector. */
template<typename T>
inline
ptr_vector<T>::ptr_vector()
{ }

/** Destructor. Deletes all objects contained in the vector. */
template<typename T>
ptr_vector<T>::~ptr_vector()
{
    delete_all();
}

/** Delete all elements. */
template<typename T>
void
ptr_vector<T>::delete_all()
{
    // delete backwards.
    for (size_type i = data.size(); i != 0; --i) {
        delete static_cast<T*>(data[i-1]);
        data[i-1] = 0;
    }
}

/** Add pointer at end. The pointed object becomes owned by this
    vector. If this function throws (due to lack of memory to expand
    the vector), it will delete \c t.
    \param t object to add */
template<typename T>
T*
ptr_vector<T>::push_back(T* t)
{
    try {
        data.push_back(t);
        return t;
    }
    catch (...) {
        delete t;
        throw;
    }
}

/** Remove last object. This will also delete the object.
    \pre !empty() */
template<typename T>
void
ptr_vector<T>::pop_back()
{
    ASSERT(!data.empty());
    delete static_cast<T*>(data.back());
    data.back() = 0;
    data.pop_back();
}

/** Swap two ptr_vector's. */
template<typename T>
inline void
ptr_vector<T>::swap(ptr_vector<T>& other)
{
    data.swap(other.data);
}

/** Get size of this vector. */
template<typename T>
inline typename ptr_vector<T>::size_type
ptr_vector<T>::size() const
{
    return data.size();
}

/** Check whether vector is empty. */
template<typename T>
inline bool
ptr_vector<T>::empty() const
{
    return data.empty();
}

/** Clear this vector. */
template<typename T>
void
ptr_vector<T>::clear()
{
    delete_all();
    data.clear();
}

/** Index access.
    \param index Element index, [0,size()) */
template<typename T>
inline T*
ptr_vector<T>::operator[](size_type index) const
{
    return static_cast<T*>(data[index]);
}

/** Get first object. */
template<typename T>
inline T*
ptr_vector<T>::front() const
{
    return static_cast<T*>(data.front());
}

/** Get last object. */
template<typename T>
inline T*
ptr_vector<T>::back() const
{
    return static_cast<T*>(data.back());
}

/** Get begin iterator. */
template<typename T>
inline typename ptr_vector<T>::const_iterator
ptr_vector<T>::begin() const
{
    return data.begin();
}

/** Get end iterator. */
template<typename T>
inline typename ptr_vector<T>::const_iterator
ptr_vector<T>::end() const
{
    return data.end();
}

/** Reserve space. Make sure we can store at least \c size objects in this
    vector without re-allocation. */
template<typename T>
inline void
ptr_vector<T>::reserve(size_type size)
{
    data.reserve(size);
}

/** Erase single element.
    \param i Iterator pointing to element. The element will be deleted. */
template<typename T>
void
ptr_vector<T>::erase(iterator i)
{
    delete *i;

    // The iterators is a vector<>::const_iterator, convert it to a regular iterator:
    typename std::vector<void*>::iterator ii = i - begin() + data.begin();
    data.erase(ii);
}

/** Erase range.
    \param i,j Range [i,j). Elements will be deleted and removed from the vector. */
template<typename T>
void
ptr_vector<T>::erase(iterator i, iterator j)
{
    for (iterator z = i; z != j; ++z)
        delete *z;

    // The iterators are vector<>::const_iterator, convert them to regular iterators:
    size_type iindex = i - begin();
    size_type jindex = j - begin();
    data.erase(data.begin() + iindex, data.begin() + jindex);

    // This used to be
    //    typename std::vector<void*>::iterator ii = i - begin() + data.begin();
    //    typename std::vector<void*>::iterator jj = j - begin() + data.begin();
    //    data.erase(ii, jj);
    // but this is miscompiled for ptr_vector<GOutbox::Message> by g++-3.3
    // (it apparently turns a '-' into a '+').
}

/** Swap elements.
    \param a first index
    \param b second index */
template<typename T>
inline void
ptr_vector<T>::swap_elems(size_type a, size_type b)
{
    std::swap(data[a], data[b]);
}

/** Swap elements of different ptr_vector's.
    \param a first index, pointing into this vector
    \param other other vector
    \param b other index, pointing into other vector */
template<typename T>
inline void
ptr_vector<T>::swap_elems(size_type a, ptr_vector& other, size_type b)
{
    std::swap(data[a], other.data[b]);
}

/** Replace an element. The replaced element will be deleted.
    \param a index
    \param value value to store there */
template<typename T>
void
ptr_vector<T>::replace_elem(size_type a, T* value)
{
    delete static_cast<T*>(data[a]);
    data[a] = value;
}

/** Destructively extract element from this vector. The element will
    be replaced by 0, the old value will be returned. The caller assumes
    responsibility for freeing it.
    \param a index */
template<typename T>
inline T*
ptr_vector<T>::extract_elem(size_type a)
{
    T* value = static_cast<T*>(data[a]);
    data[a] = 0;
    return value;
}

/** Destructively extract last element from this vector. The element
    will be popped off the vector and returned. The caller assumes
    responsibility for freeing it. */
template<typename T>
inline T*
ptr_vector<T>::extract_last()
{
    T* value = static_cast<T*>(data.back());
    data.pop_back();
    return value;
}

#endif
